import multiprocessing as mp

import numpy as np
from BoltzTraP2.fite import BOLTZMANN, FFTc, FFTev

from amset.constants import defaults
from amset.electronic_structure.fd import dfdde, fd


def get_bands_fft(
    equivalences,
    coeffs,
    lattvec,
    return_effective_mass=False,
    nworkers=defaults["nworkers"],
):
    """Rebuild the full energy bands from the interpolation coefficients.

    Args:
        equivalences: list of k-point equivalence classes in direct coordinates
        coeffs: interpolation coefficients
        lattvec: lattice vectors of the system
        return_effective_mass: Whether to calculate the effective mass.
        nworkers: number of working processes to span

    Returns:
        A 3-tuple (eband, vvband, cband): energy bands, v x v outer product
        of the velocities, and curvature of the bands (if requested). The
        shapes of those arrays are (nbands, nkpoints), (nbands, 3, 3, nkpoints)
        and (nbands, 3, 3, 3, nkpoints), where nkpoints is the total number of
        k points on the grid. If curvature is None, so will the third element
        of the tuple.
    """
    dallvec = np.vstack(equivalences)
    sallvec = mp.sharedctypes.RawArray("d", dallvec.shape[0] * 3)
    allvec = np.frombuffer(sallvec)
    allvec.shape = (-1, 3)
    dims = 2 * np.max(np.abs(dallvec), axis=0) + 1
    np.matmul(dallvec, lattvec.T, out=allvec)
    eband = np.zeros((len(coeffs), np.prod(dims)))
    vvband = np.zeros((len(coeffs), 3, 3, np.prod(dims)))
    vb = np.zeros((len(coeffs), 3, np.prod(dims)))
    if return_effective_mass:
        effective_mass = np.zeros((len(coeffs), 3, 3, np.prod(dims)))
    else:
        effective_mass = None

    # Span as many worker processes as needed, put all the bands in the queue,
    # and let them work until all the required FFTs have been computed.
    workers = []
    iqueue = mp.Queue()
    oqueue = mp.Queue()
    for iband, bandcoeff in enumerate(coeffs):
        iqueue.put((iband, bandcoeff))
    # The "None"s at the end of the queue signal the workers that there are
    # no more jobs left and they must therefore exit.
    for i in range(nworkers):
        iqueue.put(None)
    for i in range(nworkers):
        workers.append(
            mp.Process(
                target=fft_worker,
                args=(
                    equivalences,
                    sallvec,
                    dims,
                    iqueue,
                    oqueue,
                    return_effective_mass,
                ),
            )
        )
    for w in workers:
        w.start()
    # The results of the FFTs are processed as soon as they are ready.
    for r in range(len(coeffs)):
        iband, eband[iband], vvband[iband], cb, vb[iband] = oqueue.get()
        if return_effective_mass:
            effective_mass[iband] = cb
    for w in workers:
        w.join()
    if effective_mass is not None:
        effective_mass = effective_mass.real
    return eband.real, vvband.real, effective_mass, vb


def fft_worker(
    equivalences, sallvec, dims, iqueue, oqueue, return_effective_mass=False
):
    """Thin wrapper around FFTev and FFTc to be used as a worker function.

    Args:
        equivalences: list of k-point equivalence classes in direct coordinates
        sallvec: Cartesian coordinates of all k points as a 1D vector stored
                    in shared memory.
        dims: upper bound on the dimensions of the k-point grid
        iqueue: input multiprocessing.Queue used to read bad indices
            and coefficients.
        oqueue: output multiprocessing.Queue where all results of the
            interpolation are put. Each element of the queue is a 4-tuple
            of the form (index, eband, vvband, cband), containing the band
            index, the energies, the v x v outer product and the curvatures
            if requested.
        return_effective_mass: Whether to calculate the effective mass.

    Returns:
        None. The results of the calculation are put in oqueue.
    """
    iu0 = np.triu_indices(3)
    il1 = np.tril_indices(3, -1)
    iu1 = np.triu_indices(3, 1)
    allvec = np.frombuffer(sallvec)
    allvec.shape = (-1, 3)

    while True:
        task = iqueue.get()
        if task is None:
            break
        else:
            index, bandcoeff = task
        eband, vb = FFTev(equivalences, bandcoeff, allvec, dims)
        vvband = np.zeros((3, 3, np.prod(dims)))
        effective_mass = np.zeros((3, 3, np.prod(dims)))

        vvband[iu0[0], iu0[1]] = vb[iu0[0]] * vb[iu0[1]]
        vvband[il1[0], il1[1]] = vvband[iu1[0], iu1[1]]
        if return_effective_mass:
            effective_mass[iu0] = FFTc(equivalences, bandcoeff, allvec, dims)
            effective_mass[il1] = effective_mass[iu1]
            effective_mass = np.linalg.inv(effective_mass.T).T
        else:
            effective_mass = None
        oqueue.put((index, eband, vvband, effective_mass, vb))


def fermiintegrals(epsilon, dos, sigma, mur, Tr, dosweight=2.0, cdos=None):
    """Compute the moments of the FD distribution over the band structure.

    Args:
        epsilon: array of energies at which the DOS is available
        dos: density of states
        sigma: transport DOS
        mur: array of chemical potential values
        Tr: array of temperature values
        dosweight: maximum occupancy of an electron mode
        cdos: "curvature DOS" if available

    Returns:
        Five numpy arrays, namely:
        1. An (nT, nmu) array with the electron counts for each temperature and
           each chemical potential.
        2. An (nT, nmu, 3, 3) with the integrals of the 3 x 3 transport DOS
           over the band structure taking the occupancies into account.
        3. An (nT, nmu, 3, 3) with the first moment of the 3 x 3 transport DOS
           over the band structure taking the occupancies into account.
        4. An (nT, nmu, 3, 3) with the second moment of the 3 x 3 transport DOS
           over the band structure taking the occupancies into account.
        5. If the cdos argument is provided, an (nT, nmu, 3, 3, 3) with the
           integrals of the 3 x 3 x 3 "curvature DOS" over the band structure
           taking the occupancies into account.
        where nT and nmu are the sizes of Tr and mur, respectively.
    """
    kBTr = np.array(Tr) * BOLTZMANN
    nT = len(Tr)
    nmu = len(mur)
    N = np.empty((nT, nmu))
    L0 = np.empty((nT, nmu, 3, 3))
    L1 = np.empty((nT, nmu, 3, 3))
    L2 = np.empty((nT, nmu, 3, 3))
    if cdos is not None:
        L11 = np.empty((nT, nmu, 3, 3, 3))
    else:
        L11 = None
    de = epsilon[1] - epsilon[0]
    for iT, kBT in enumerate(kBTr):
        for imu, mu in enumerate(mur):
            N[iT, imu] = -(dosweight * dos * fd(epsilon, mu, kBT)).sum() * de
            int0 = -dosweight * dfdde(epsilon, mu, kBT)
            intn = int0 * sigma
            L0[iT, imu] = intn.sum(axis=2) * de
            intn *= epsilon - mu
            L1[iT, imu] = -intn.sum(axis=2) * de
            intn *= epsilon - mu
            L2[iT, imu] = intn.sum(axis=2) * de
            if cdos is not None:
                cint = int0 * cdos
                L11[iT, imu] = -cint.sum(axis=3) * de
    return N, L0, L1, L2, L11



import warnings

import matplotlib.pyplot
import numpy as np
from matplotlib import cm

from amset.plot import (
    BaseTransportPlotter,
    amset_base_style,
    get_figsize,
    pretty_subplot,
    styled_plot,
)

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

from amset.plot.base import PlotData

property_data = {
    "conductivity": {
        "label": "Conductivity (S/m)",
        "label_symbol": r"$\sigma$ (S/m)",
        "log": True,
    },
    "seebeck": {
        "label": r"Seebeck coefficient ($\mu$V/K)",
        "label_symbol": r"$S$ ($\mu$V/K)",
        "log": False,
    },
    "thermal conductivity": {
        "label": "Thermal conductivity (W/mK)",
        "label_symbol": r"$\kappa_e$ (W/mK)",
        "log": True,
    },
    "mobility": {
        "label": "Mobility (cm$^2$/Vs)",
        "label_symbol": r"$\mu$ (cm$^2$/Vs)",
        "log": False,
    },
    "power factor": {
        "label": r"Power factor (mW$\,$m$^{-1}$$\,$K$^{-2}$)",
        "label_symbol": r"PF (mW$\,$m$^{-1}$$\,$K$^{-2}$)",
        "log": False,
    },
    "temperature": {
        "label": "Temperature (K)",
        "label_symbol": "$T$ (K)",
        "log": False,
    },
    "doping": {
        "label": "Carrier concentration (cm$^{-3}$)",
        "label_symbol": "$n$ (cm$^{-3}$)",
        "log": True,
    },
}
dir_mapping = {0: "xx", 1: "yy", 2: "zz"}
ls_mapping = {0: "-", 1: "--", 2: ":"}


class TransportPlotter(BaseTransportPlotter):
    def __init__(self, data, use_symbol=False, pad=0.05, average=True):
        super().__init__(data)
        self.label_key = "label_symbol" if use_symbol else "label"
        self.pad = pad
        self.average = average

    @styled_plot(amset_base_style)
    def get_plot(
        self,
        properties=("conductivity", "seebeck", "thermal conductivity"),
        x_property=None,
        doping_idx=None,
        temperature_idx=None,
        grid=None,
        height=None,
        width=None,
        doping_type=None,
        conductivity_label=None,
        conductivity_min=None,
        conductivity_max=None,
        log_conductivity=None,
        seebeck_label=None,
        seebeck_min=None,
        seebeck_max=None,
        log_seebeck=None,
        mobility_label=None,
        mobility_min=None,
        mobility_max=None,
        log_mobility=None,
        thermal_conductivity_label=None,
        thermal_conductivity_min=None,
        thermal_conductivity_max=None,
        log_thermal_conductivity=None,
        power_factor_label=None,
        power_factor_min=None,
        power_factor_max=None,
        log_power_factor=None,
        xlabel=None,
        xmin=None,
        xmax=None,
        logx=None,
        legend=True,
        axes=None,
        return_plot_data=False,
        plt=None,
        style=None,
        no_base_style=False,
        fonts=None,
    ):

        if axes is None:
            if grid is None or len(grid) == 0:
                grid = (1, len(properties))

            if np.product(grid) < len(properties):
                raise ValueError("Grid is not large enough to plot all properties")

            width, height = get_figsize(*grid, width=width, height=height)
            _, axes = pretty_subplot(*grid, width=width, height=height)
        else:
            if not isinstance(axes, (list, tuple, np.ndarray)):
                # single axis given
                grid = (1, 1)
            else:
                axes = np.array(axes)
                if grid is None:
                    grid = axes.shape
                    if len(grid) == 1:
                        grid = (1, grid[0])

            if np.product(grid) < len(properties):
                raise ValueError("Not enough axes to plot all properties")

        if doping_type is not None and doping_idx is not None:
            warnings.warn(
                "Both doping type and doping indexes have been set. This can cause "
                "unexpected behaviour."
            )

        prop_dict = {
            "conductivity": {
                "ylabel": conductivity_label,
                "ymin": conductivity_min,
                "ymax": conductivity_max,
                "logy": log_conductivity,
            },
            "mobility": {
                "ylabel": mobility_label,
                "ymin": mobility_min,
                "ymax": mobility_max,
                "logy": log_mobility,
            },
            "seebeck": {
                "ylabel": seebeck_label,
                "ymin": seebeck_min,
                "ymax": seebeck_max,
                "logy": log_seebeck,
            },
            "thermal conductivity": {
                "ylabel": thermal_conductivity_label,
                "ymin": thermal_conductivity_min,
                "ymax": thermal_conductivity_max,
                "logy": log_thermal_conductivity,
            },
            "power factor": {
                "ylabel": power_factor_label,
                "ymin": power_factor_min,
                "ymax": power_factor_max,
                "logy": log_power_factor,
            },
        }

        if np.product(grid) > len(properties):
            n_missing = np.product(grid) - len(properties)
            properties = list(properties) + [None] * n_missing

        plot_data = []
        for (x, y), prop in zip(np.ndindex(grid), properties):
            if grid[0] == 1 and grid[1] == 1:
                ax = axes
            elif grid[0] == 1:
                ax = axes[y]
            elif grid[1] == 1:
                ax = axes[x]
            else:
                ax = axes[x, y]

            if prop is None:
                ax.axis("off")

            else:
                prop_data = self.plot_property(
                    ax,
                    prop,
                    x_property=x_property,
                    temperature_idx=temperature_idx,
                    doping_idx=doping_idx,
                    doping_type=doping_type,
                    xmin=xmin,
                    xmax=xmax,
                    xlabel=xlabel,
                    logx=logx,
                    **prop_dict[prop],
                )
                plot_data.append(prop_data)

        if legend:
            if grid[0] == 1 and grid[1] == 1:
                ax = axes
            elif grid[0] == 1:
                ax = axes[-1]
            elif grid[1] == 1:
                ax = axes[0]
            else:
                ax = axes[0, -1]
            ax.legend(loc="upper left", bbox_to_anchor=(1.01, 1), ncol=1)

        if return_plot_data:
            return matplotlib.pyplot, plot_data

        return matplotlib.pyplot

    def plot_property(
        self,
        ax,
        prop,
        x_property=None,
        temperature_idx=None,
        doping_idx=None,
        doping_type=None,
        ylabel=None,
        ymin=None,
        ymax=None,
        logy=None,
        xlabel=None,
        xmin=None,
        xmax=None,
        logx=None,
    ):
        if temperature_idx is None:
            temperature_idx = np.arange(self.temperatures.shape[0])
        elif isinstance(temperature_idx, int):
            temperature_idx = [temperature_idx]
        temperatures = self.temperatures[temperature_idx]

        if doping_idx is None:
            doping_idx = np.arange(self.doping.shape[0])
        elif isinstance(doping_idx, int):
            doping_idx = [doping_idx]

        if doping_type == "n":
            doping_idx = [i for i in doping_idx if self.doping[i] <= 0]
        elif doping_type == "p":
            doping_idx = [i for i in doping_idx if self.doping[i] >= 0]

        doping = self.doping[doping_idx]

        data = self._get_data(prop)

        if x_property is None and len(temperatures) == 1:
            x_property = "doping"
        elif x_property is None:
            x_property = "temperature"

        y = data[doping_idx][:, temperature_idx]
        if self.average:
            y = y.reshape((len(doping), len(temperatures)))  # make sure y is still 2D
        else:
            y = y.reshape((len(doping), len(temperatures), 3))
        if x_property == "doping":
            labels = [fancy_format_temp(t) for t in temperatures]
            x = doping
            y = np.swapaxes(y, 0, 1)
        elif x_property == "temperature":
            labels = [fancy_format_doping(d) for d in doping]
            x = temperatures
            y = y
        else:
            raise ValueError(f"Unknown x_property: {x_property}")

        if np.any((x > 0) & (x < 0)):
            raise ValueError(
                "Cannot plot n- and p- type properties on the same figure."
                "Use doping-idx to select doping indices"
            )
        x = np.abs(x)

        cmap = cm.get_cmap("viridis_r")
        colors = cmap(np.linspace(0.03, 1 - 0.03, len(y)))

        line_data = []
        for yi, label, color in zip(y, labels, colors):
            if self.average:
                lines = ax.plot(x, yi, label=label, c=color)
                line_data.extend(lines)
            else:
                for i in range(3):
                    lines = ax.plot(
                        x,
                        yi[:, i],
                        label=f"{dir_mapping[i]} {label}",
                        c=color,
                        ls=ls_mapping[i],
                    )
                    line_data.extend(lines)

        ylabel = ylabel if ylabel else property_data[prop][self.label_key]
        xlabel = xlabel if xlabel else property_data[x_property][self.label_key]
        logy = logy if logy is not None else property_data[prop]["log"]
        logx = logx if logx is not None else property_data[x_property]["log"]
        ylim = get_lim(y, ymin, ymax, logy, self.pad)
        xlim = get_lim(x, xmin, xmax, logx, self.pad)

        if logy:
            ax.semilogy()
        if logx:
            ax.semilogx()
        ax.set(ylabel=ylabel, xlabel=xlabel, ylim=ylim, xlim=xlim)

        return PlotData(
            x_property=x_property,
            y_property=prop,
            x_label=xlabel,
            y_label=ylabel,
            labels=[line.get_label() for line in line_data],
            x=x,
            y=np.array([line.get_ydata() for line in line_data]),
        )

    def _get_data(self, prop):
        if prop == "conductivity":
            data = self.conductivity
        elif prop == "seebeck":
            data = self.seebeck
        elif prop == "thermal conductivity":
            data = self.electronic_thermal_conductivity
        elif prop == "mobility":
            data = self.mobility
        elif prop == "power factor":
            data = self.seebeck ** 2 * self.conductivity * 1e-9  # convert to mW/(m K^2)
        else:
            raise ValueError(f"Unrecognised property: {prop}")

        if self.average:
            data = np.average(np.linalg.eigvalsh(data), axis=-1)
        else:
            data = np.diagonal(data, axis1=-2, axis2=-1)
        return data


def get_lim(data, vmin, vmax, logv, pad):
    data = data[np.isfinite(data)]
    data_min = np.log10(data.min()) if logv else data.min()
    data_max = np.log10(data.max()) if logv else data.max()
    data_pad = pad * (data_max - data_min)

    if vmin is None:
        vmin = data_min - data_pad
        if logv:
            vmin = 10 ** vmin

    if vmax is None:
        vmax = data_max + data_pad
        if logv:
            vmax = 10 ** vmax

    return vmin, vmax


def carrier_type(doping):
    if doping < 0:
        return "n"
    else:
        return "p"


def fancy_format_doping(doping):
    if carrier_type(doping) == "n":
        doping = abs(doping)
        prefix = "$n_e$"
    else:
        prefix = "$n_h$"
    return f"{prefix} = {format_doping(doping)}"


def fancy_format_temp(temperature):
    return f"T = {format_temp(temperature)}"


def format_temp(temperature):
    if temperature % 10 == 0:
        return f"{int(temperature)} K"
    else:
        return f"{temperature:.1f} K"


def format_doping(doping):
    log_doping = np.log10(doping)
    if log_doping % 1 == 0:
        return f"10$^{{{int(log_doping)}}}$ cm$^{{-3}}$"
    else:
        log_part = int(np.floor(log_doping))
        front_part = 10 ** (log_doping - log_part)
        return fr"{front_part:.2f}$\times$10$^{{{log_part}}}$ cm$^{{-3}}$"



import logging

import numpy as np
from pymatgen.electronic_structure.bandstructure import BandStructure
from pymatgen.electronic_structure.core import Spin
from pymatgen.util.coord import pbc_diff

from amset.constants import defaults
from amset.electronic_structure.common import get_ibands, get_vb_idx
from amset.electronic_structure.symmetry import expand_kpoints
from amset.interpolation.periodic import (
    PeriodicLinearInterpolator,
    group_bands_and_kpoints,
)

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)


class ProjectionOverlapCalculator(PeriodicLinearInterpolator):
    def __init__(self, nbands, data_shape, interpolators, rotation_mask, band_centers):
        super().__init__(nbands, data_shape, interpolators)
        self.rotation_mask = rotation_mask
        self.band_centers = band_centers

    def to_reference(self):
        interpolator_references = self._interpolators_to_reference()
        return (
            self.nbands,
            self.data_shape,
            interpolator_references,
            self.rotation_mask,
            self.band_centers,
        )

    @classmethod
    def from_data(
        cls, kpoints, data, rotation_mask=None, band_centers=None, gaussian=False
    ):
        logger.info("Initializing orbital overlap calculator")
        if rotation_mask is None or band_centers is None:
            raise ValueError("rotation_mask and band_centers are both required.")

        grid_kpoints, mesh_dim, sort_idx = cls._grid_kpoints(kpoints)
        nbands, data_shape, interpolators = cls._setup_interpolators(
            data, grid_kpoints, mesh_dim, sort_idx, gaussian
        )
        return cls(nbands, data_shape, interpolators, rotation_mask, band_centers)

    @classmethod
    def from_band_structure(
        cls,
        band_structure: BandStructure,
        energy_cutoff=defaults["energy_cutoff"],
        symprec=defaults["symprec"],
    ):
        kpoints = np.array([k.frac_coords for k in band_structure.kpoints])
        efermi = band_structure.efermi
        structure = band_structure.structure

        full_kpoints, _, _, _, _, ir_to_full_idx = expand_kpoints(
            structure, kpoints, symprec=symprec, return_mapping=True, time_reversal=True
        )

        ibands = get_ibands(energy_cutoff, band_structure)
        vb_idx = get_vb_idx(energy_cutoff, band_structure)

        energies = {s: e[ibands[s]] for s, e in band_structure.bands.items()}
        energies = {s: e[:, ir_to_full_idx] for s, e in energies.items()}
        projections = {s: p[ibands[s]] for s, p in band_structure.projections.items()}

        band_centers = get_band_centers(full_kpoints, energies, vb_idx, efermi)
        rotation_mask = get_rotation_mask(projections)

        full_projections = {}
        for spin, spin_projections in projections.items():
            nbands = spin_projections.shape[0]
            nkpoints = len(full_kpoints)

            spin_projections = spin_projections[:, ir_to_full_idx].reshape(
                (nbands, nkpoints, -1), order="F"
            )
            spin_projections /= np.linalg.norm(spin_projections, axis=2)[..., None]
            spin_projections[np.isnan(spin_projections)] = 0
            full_projections[spin] = spin_projections

        return cls.from_data(
            full_kpoints,
            full_projections,
            rotation_mask=rotation_mask,
            band_centers=band_centers,
        )

    def get_coefficients(self, spin, bands, kpoints):
        return self.interpolate(spin, bands, kpoints)

    def get_overlap(self, spin, band_a, kpoint_a, band_b, kpoint_b):
        bands, kpoints, single_overlap = group_bands_and_kpoints(
            band_a, kpoint_a, band_b, kpoint_b
        )
        p = self.get_coefficients(spin, bands, kpoints)

        centers = self.band_centers[spin][band_a]
        center = centers[np.argmin(np.linalg.norm(pbc_diff(centers, kpoint_a), axis=1))]

        shift_a = pbc_diff(kpoint_a, center)
        shift_b = pbc_diff(kpoint_b, center)

        angles = cosine(shift_a, shift_b)
        prod = p[0] * p[1:]
        overlap = (
            prod * (1 - self.rotation_mask)
            + prod * self.rotation_mask * angles[:, None]
        )
        overlap = overlap.sum(axis=1) ** 2

        if single_overlap:
            return overlap[0]
        else:
            return overlap


def get_rotation_mask(projections):
    nprojections, natoms = projections[Spin.up].shape[2:]
    mask = np.ones(nprojections)
    mask[0] = 0
    return np.tile(mask, natoms)


def cosine(v1, v2):
    # v2 can be list of vectors or single vector

    return_single = False
    if len(v2.shape) == 1:
        v2 = v2[None, :]
        return_single = True

    v1_norm = np.linalg.norm(v1)
    v2_norm = np.linalg.norm(v2, axis=1)

    v_dot = np.dot(v1, v2.T)
    v_angle = v_dot / (v1_norm * v2_norm)
    v_angle[np.isnan(v_angle)] = 1

    if return_single:
        return v_angle[0]
    else:
        return v_angle


def get_band_centers(kpoints, energies, vb_idx, efermi, tol=0.0001):
    band_centers = {}

    for spin, spin_energies in energies.items():
        spin_centers = []
        for i, band_energies in enumerate(spin_energies):
            if vb_idx is None:
                # handle metals
                k_idxs = np.abs(band_energies - efermi) < tol

            elif i <= vb_idx[spin]:
                k_idxs = (np.max(band_energies) - band_energies) < tol

            else:
                k_idxs = (band_energies - np.min(band_energies)) < tol

            if len(k_idxs) > 0:
                k_idxs = [0]

            spin_centers.append(kpoints[k_idxs])
        band_centers[spin] = spin_centers
    return band_centers



import logging
from collections import defaultdict
from typing import Dict, Optional, Tuple, Union

import numpy as np
from matplotlib import rcParams
from matplotlib.axes import SubplotBase
from matplotlib.axis import Axis
from matplotlib.colors import LogNorm
from matplotlib.ticker import AutoMinorLocator, MaxNLocator
from pymatgen.electronic_structure.bandstructure import (
    BandStructure,
    BandStructureSymmLine,
)
from pymatgen.electronic_structure.core import Spin
from pymatgen.electronic_structure.plotter import BSPlotter
from sumo.plotting import pretty_plot
from sumo.symmetry import Kpath, PymatgenKpath

from amset.constants import defaults, hbar
from amset.interpolation.bandstructure import Interpolator
from amset.interpolation.periodic import PeriodicLinearInterpolator
from amset.log import initialize_amset_logger
from amset.plot import BaseMeshPlotter, amset_base_style, styled_plot

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)


class LineshapePlotter(BaseMeshPlotter):
    def __init__(
        self,
        data,
        interpolation_factor=5,
        print_log=defaults["print_log"],
        symprec=defaults["symprec"],
    ):
        super().__init__(data)
        self.interpolation_factor = interpolation_factor

        if print_log:
            initialize_amset_logger(filename="lineshape.log")
        self.symprec = symprec

    def _get_interpolater(self, n_idx, t_idx, mode="linear"):
        props = defaultdict(dict)
        for spin in self.spins:
            # calculate total rate
            spin_rates = np.sum(self.scattering_rates[spin][:, n_idx, t_idx], axis=0)

            # easier to interpolate the log
            log_rates = np.log10(spin_rates)

            # # handle rates that close to numerical noise
            log_rates[log_rates > 18] = 15
            log_rates[np.isnan(log_rates)] = 15

            # map to full k-point mesh
            props[spin]["rates"] = log_rates

        if mode == "linear":
            return _LinearBandStructureInterpolator(
                self.kpoints,
                self.ir_to_full_kpoint_mapping,
                self.energies,
                self.structure,
                self.efermi,
                props,
            )
        elif mode == "fourier":
            bs = BandStructure(
                self.ir_kpoints,
                self.energies,
                self.structure.lattice,
                self.efermi,
                structure=self.structure,
            )

            return Interpolator(
                bs,
                self.num_electrons,
                interpolation_factor=self.interpolation_factor,
                soc=self.soc,
                other_properties=props,
            )
        raise ValueError("Unknown interpolation mode; should be 'linear' or 'fourier'.")

    @styled_plot(amset_base_style)
    def get_plot(
        self,
        n_idx,
        t_idx,
        zero_to_efermi=True,
        estep=0.01,
        line_density=100,
        height=3.2,
        width=3.2,
        emin=None,
        emax=None,
        amin=5e-5,
        amax=1e-1,
        ylabel="Energy (eV)",
        plt=None,
        aspect=None,
        kpath=None,
        cmap="viridis",
        colorbar=True,
        style=None,
        no_base_style=False,
        fonts=None,
    ):
        interpolater = self._get_interpolater(n_idx, t_idx)

        bs, prop = interpolater.get_line_mode_band_structure(
            line_density=line_density,
            return_other_properties=True,
            kpath=kpath,
            symprec=self.symprec,
        )
        bs, rates = force_branches(bs, {s: p["rates"] for s, p in prop.items()})

        fd_emin, fd_emax = self.fd_cutoffs
        if not emin:
            emin = fd_emin
            if zero_to_efermi:
                emin -= bs.efermi

        if not emax:
            emax = fd_emax
            if zero_to_efermi:
                emax -= bs.efermi

        logger.info("Plotting band structure")
        if isinstance(plt, (Axis, SubplotBase)):
            ax = plt
        else:
            plt = pretty_plot(width=width, height=height, plt=plt)
            ax = plt.gca()

        if zero_to_efermi:
            bs.bands = {s: b - bs.efermi for s, b in bs.bands.items()}
            bs.efermi = 0

        bs_plotter = BSPlotter(bs)
        plot_data = bs_plotter.bs_plot_data(zero_to_efermi=zero_to_efermi)

        energies = np.linspace(emin, emax, int((emax - emin) / estep))
        distances = np.array([d for x in plot_data["distances"] for d in x])

        # rates are currently log(rate)
        mesh_data = np.full((len(distances), len(energies)), 0.0)
        for spin in self.spins:
            for spin_energies, spin_rates in zip(bs.bands[spin], rates[spin]):
                for d_idx in range(len(distances)):
                    energy = spin_energies[d_idx]
                    linewidth = 10 ** spin_rates[d_idx] * hbar / 2
                    broadening = lorentzian(energies, energy, linewidth)
                    broadening /= 1000  # convert 1/eV to 1/meV
                    mesh_data[d_idx] += broadening

        im = ax.pcolormesh(
            distances,
            energies,
            mesh_data.T,
            rasterized=True,
            cmap=cmap,
            norm=LogNorm(vmin=amin, vmax=amax),
            shading="auto",
        )
        if colorbar:
            pos = ax.get_position()
            cax = plt.gcf().add_axes([pos.x1 + 0.035, pos.y0, 0.035, pos.height])
            cbar = plt.colorbar(im, cax=cax)
            cbar.ax.tick_params(axis="y", length=rcParams["ytick.major.size"] * 0.5)
            cbar.ax.set_ylabel(
                r"$A_\mathbf{k}$ (meV$^{-1}$)", rotation=270, va="bottom"
            )

        _maketicks(ax, bs_plotter, ylabel=ylabel)
        _makeplot(
            ax,
            plot_data,
            bs,
            zero_to_efermi=zero_to_efermi,
            width=width,
            height=height,
            ymin=emin,
            ymax=emax,
            aspect=aspect,
        )
        return plt


def _makeplot(
    ax,
    data,
    bs,
    zero_to_efermi=True,
    ymin=-3.0,
    ymax=3.0,
    height=None,
    width=None,
    aspect=None,
):
    """Tidy the band structure & add the density of states if required."""
    # draw line at Fermi level if not zeroing to e-Fermi
    if not zero_to_efermi:
        ytick_color = rcParams["ytick.color"]
        ef = bs.efermi
        ax.axhline(ef, color=ytick_color)

    # set x and y limits
    ax.set_xlim(0, data["distances"][-1][-1])
    if bs.is_metal() and not zero_to_efermi:
        ax.set_ylim(bs.efermi + ymin, bs.efermi + ymax)
    else:
        ax.set_ylim(ymin, ymax)

    # keep correct aspect ratio for axes based on canvas size
    if aspect is not False:
        x0, x1 = ax.get_xlim()
        y0, y1 = ax.get_ylim()
        if width is None:
            width = rcParams["figure.figsize"][0]
        if height is None:
            height = rcParams["figure.figsize"][1]

        if not aspect:
            aspect = height / width

        ax.set_aspect(aspect * ((x1 - x0) / (y1 - y0)))


def _maketicks(ax, bs_plotter, ylabel="Energy (eV)"):
    """Utility method to add tick marks to a band structure."""
    # set y-ticks
    ax.yaxis.set_major_locator(MaxNLocator(6))
    ax.yaxis.set_minor_locator(AutoMinorLocator(2))

    # set x-ticks; only plot the unique tick labels
    ticks = bs_plotter.get_ticks()
    unique_d = []
    unique_l = []
    if ticks["distance"]:
        temp_ticks = list(zip(ticks["distance"], ticks["label"]))
        unique_d.append(temp_ticks[0][0])
        unique_l.append(temp_ticks[0][1])
        for i in range(1, len(temp_ticks)):
            # Append label to sequence if it is not same as predecessor
            if unique_l[-1] != temp_ticks[i][1]:
                unique_d.append(temp_ticks[i][0])
                unique_l.append(temp_ticks[i][1])

    logging.info("Label positions:")
    for dist, label in list(zip(unique_d, unique_l)):
        logging.info(f"\t{dist:.4f}: {label}")

    ax.set_xticks(unique_d)
    ax.set_xticklabels(unique_l)
    ax.xaxis.grid(True)
    ax.set_ylabel(ylabel)


def lorentzian(x, x0, gamma):
    return 1 / np.pi * gamma / ((x - x0) ** 2 + gamma ** 2)


class _LinearBandStructureInterpolator:
    def __init__(
        self,
        full_kpoints,
        ir_to_full_idx,
        energies,
        structure,
        efermi,
        other_properties,
    ):
        self.structure = structure
        self.efermi = efermi
        self.spins = list(energies.keys())
        self.nbands = {s: len(e) for s, e in energies.items()}

        full_energies = {s: e[:, ir_to_full_idx] for s, e in energies.items()}
        self.bs_interpolator = PeriodicLinearInterpolator.from_data(
            full_kpoints, full_energies
        )

        self.property_interpolators = {}
        other_properties = _transpose_dict(other_properties)
        for prop, prop_data in other_properties.items():

            full_prop_data = {s: p[:, ir_to_full_idx] for s, p in prop_data.items()}
            self.property_interpolators[prop] = PeriodicLinearInterpolator.from_data(
                full_kpoints, full_prop_data, gaussian=0.75
            )

    def get_line_mode_band_structure(
        self,
        line_density: int = 50,
        kpath: Optional[Kpath] = None,
        symprec: Optional[float] = defaults["symprec"],
        return_other_properties: bool = False,
    ) -> Union[
        BandStructureSymmLine,
        Tuple[BandStructureSymmLine, Dict[Spin, Dict[str, np.ndarray]]],
    ]:
        """Gets the interpolated band structure along high symmetry directions.

        Args:
            line_density: The maximum number of k-points between each two
                consecutive high-symmetry k-points
            symprec: The symmetry tolerance used to determine the space group
                and high-symmetry path.
            return_other_properties: Whether to include the interpolated
                other_properties data for each k-point along the band structure path.

        Returns:
            The line mode band structure.
        """
        if not kpath:
            kpath = PymatgenKpath(self.structure, symprec=symprec)

        kpoints, labels = kpath.get_kpoints(line_density=line_density, cart_coords=True)
        labels_dict = {
            label: kpoint for kpoint, label in zip(kpoints, labels) if label != ""
        }

        rlat = self.structure.lattice.reciprocal_lattice
        frac_kpoints = rlat.get_fractional_coords(kpoints)

        energies = {}
        other_properties = defaultdict(dict)
        for spin in self.spins:
            energies[spin] = self._interpolate_spin(
                spin, frac_kpoints, self.bs_interpolator
            )

            if return_other_properties:
                for prop, property_interpolator in self.property_interpolators.items():
                    other_properties[spin][prop] = self._interpolate_spin(
                        spin, frac_kpoints, property_interpolator
                    )

        bs = BandStructureSymmLine(
            kpoints,
            energies,
            rlat,
            self.efermi,
            labels_dict,
            coords_are_cartesian=True,
            structure=self.structure,
        )

        if return_other_properties:
            return bs, other_properties
        else:
            return bs

    def _interpolate_spin(self, spin, kpoints, interpolator):
        nkpoints = len(kpoints)
        spin_nbands = self.nbands[spin]
        ibands = np.repeat(np.arange(spin_nbands), nkpoints)
        all_kpoints = np.tile(kpoints, (spin_nbands, 1))
        data = interpolator.interpolate(spin, ibands, all_kpoints)
        return data.reshape(spin_nbands, nkpoints)


def _transpose_dict(d):
    td = defaultdict(dict)
    for k1, v1 in d.items():
        for k2, v2 in v1.items():
            td[k2][k1] = v2
    return td


def force_branches(bandstructure, other_property=None):
    """Force a linemode band structure to contain branches.

    Branches give a specific portion of the path from one high-symmetry point
    to another. Branches are required for the plotting methods to function correctly.
    Unfortunately, due to the pymatgen BandStructure implementation they require
    duplicate k-points in the band structure path. To avoid this unnecessary
    computational expense, this function can reconstruct branches in band structures
    without the duplicate k-points.

    Args:
        bandstructure: A band structure object.
        other_property: Another property with the format {spin: (nbands, nkpts, ...)
            to split into branches.

    Returns:
        A band structure with brnaches.
    """
    kpoints = np.array([k.frac_coords for k in bandstructure.kpoints])
    labels_dict = {k: v.frac_coords for k, v in bandstructure.labels_dict.items()}

    # pymatgen band structure objects support branches. These are formed when
    # two kpoints with the same label are next to each other. This bit of code
    # will ensure that the band structure will contain branches, if it doesn't
    # already.
    dup_ids = []
    high_sym_kpoints = tuple(map(tuple, labels_dict.values()))
    for i, k in enumerate(kpoints):
        dup_ids.append(i)
        if (
            tuple(k) in high_sym_kpoints
            and i != 0
            and i != len(kpoints) - 1
            and (
                not np.array_equal(kpoints[i + 1], k)
                or not np.array_equal(kpoints[i - 1], k)
            )
        ):
            dup_ids.append(i)

    kpoints = kpoints[dup_ids]

    eigenvals = {}
    projections = {}
    for spin, spin_energies in bandstructure.bands.items():
        eigenvals[spin] = spin_energies[:, dup_ids]
        if len(bandstructure.projections) != 0:
            projections[spin] = bandstructure.projections[spin][:, dup_ids]

    new_property = {}
    if other_property is not None:
        for spin, spin_prop in other_property.items():
            new_property[spin] = spin_prop[:, dup_ids]

    new_bandstructure = type(bandstructure)(
        kpoints,
        eigenvals,
        bandstructure.lattice_rec,
        bandstructure.efermi,
        labels_dict,
        structure=bandstructure.structure,
        projections=projections,
    )
    return new_bandstructure, new_property



import warnings

import matplotlib
import numpy as np

from amset.plot import get_figsize, pretty_subplot, styled_plot
from amset.plot.base import BaseTransportPlotter, amset_base_style, base_total_color
from amset.plot.transport import (
    dir_mapping,
    fancy_format_doping,
    fancy_format_temp,
    get_lim,
    ls_mapping,
)

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"


_property_data = {
    "mobility": {
        "label": "Mobility (cm$^2$/Vs)",
        "label_symbol": r"$\mu$ (cm$^2$/Vs)",
        "log": True,
    },
    "temperature": {
        "label": "Temperature (K)",
        "label_symbol": "$T$ (K)",
        "log": False,
    },
    "doping": {
        "label": "Carrier concentration (cm$^{-3}$)",
        "label_symbol": "$n$ (cm$^{-3}$)",
        "log": True,
    },
}


class MobilityPlotter(BaseTransportPlotter):
    def __init__(
        self, data, use_symbol=False, pad=0.05, average=True, separate_mobility=True
    ):
        super().__init__(data)
        self.label_key = "label_symbol" if use_symbol else "label"
        self.pad = pad
        self.average = average
        self.separate_mobility = separate_mobility

    @styled_plot(amset_base_style)
    def get_plot(
        self,
        x_property=None,
        doping_idx=None,
        temperature_idx=None,
        doping_type=None,
        grid=None,
        height=None,
        width=None,
        ylabel=None,
        ymin=None,
        ymax=None,
        logy=None,
        xlabel=None,
        xmin=None,
        xmax=None,
        logx=None,
        legend=True,
        title=True,
        total_color=None,
        axes=None,
        style=None,
        no_base_style=False,
        fonts=None,
    ):
        if doping_type is not None and doping_idx is not None:
            warnings.warn(
                "Both doping type and doping indexes have been set. This can cause "
                "unexpected behaviour."
            )

        if temperature_idx is None:
            temperature_idx = np.arange(self.temperatures.shape[0])
        elif isinstance(temperature_idx, int):
            temperature_idx = [temperature_idx]

        if doping_idx is None:
            doping_idx = np.arange(self.doping.shape[0])
        elif isinstance(doping_idx, int):
            doping_idx = [doping_idx]

        if doping_type == "n":
            doping_idx = [i for i in doping_idx if self.doping[i] <= 0]
        elif doping_type == "p":
            doping_idx = [i for i in doping_idx if self.doping[i] >= 0]

        if x_property is None and len(temperature_idx) == 1:
            x_property = "doping"
        elif x_property is None:
            x_property = "temperature"

        if x_property == "doping":
            if grid is None or len(grid) == 0:
                grid = (1, len(temperature_idx))
            primary_idxs = temperature_idx
            secondary_idxs = doping_idx
        elif x_property == "temperature":
            if grid is None or len(grid) == 0:
                grid = (1, len(doping_idx))
            primary_idxs = doping_idx
            secondary_idxs = temperature_idx
        else:
            raise ValueError(f"Unrecognised x_property: {x_property}")

        if axes is None:
            width, height = get_figsize(*grid, width=width, height=height)
            _, axes = pretty_subplot(*grid, width=width, height=height)
        else:
            if not isinstance(axes, (list, tuple, np.ndarray)):
                # single axis given
                grid = (1, 1)
            else:
                axes = np.array(axes)
                if grid is None:
                    grid = axes.shape
                    if len(grid) == 1:
                        grid = (1, grid[0])

            if len(primary_idxs) > np.product(grid):
                raise ValueError(
                    "Axes array not commensurate with doping_idx and temperature_idx"
                )

        for (x, y), primary_idx in zip(np.ndindex(grid), primary_idxs):
            if grid[0] == 1 and grid[1] == 1:
                ax = axes
            elif grid[0] == 1:
                ax = axes[y]
            elif grid[1] == 1:
                ax = axes[x]
            else:
                ax = axes[x, y]

            self.plot_mobility(
                ax,
                x_property,
                primary_idx,
                secondary_idxs,
                xlabel=xlabel,
                xmin=xmin,
                xmax=xmax,
                logx=logx,
                ylabel=ylabel,
                ymin=ymin,
                ymax=ymax,
                logy=logy,
                title=title,
                total_color=total_color,
            )
            ax.set()

        if legend and (self.separate_mobility or not self.average):
            if grid[0] == 1 and grid[1] == 1:
                ax = axes
            elif grid[0] == 1:
                ax = axes[-1]
            elif grid[1] == 1:
                ax = axes[0]
            else:
                ax = axes[0, -1]
            ax.legend(loc="upper left", bbox_to_anchor=(1.01, 1), ncol=1)

        return matplotlib.pyplot

    def plot_mobility(
        self,
        ax,
        x_property,
        primary_idx,
        secondary_idxs,
        xlabel=None,
        xmin=None,
        xmax=None,
        logx=None,
        ylabel=None,
        ymin=None,
        ymax=None,
        logy=None,
        title=True,
        total_color=None,
    ):
        if x_property == "temperature":
            labels, y = self._get_data(primary_idx, secondary_idxs)
            x = self.temperatures[secondary_idxs]
            title_str = fancy_format_doping(self.doping[primary_idx])
        elif x_property == "doping":
            labels, y = self._get_data(secondary_idxs, primary_idx)
            x = self.doping[secondary_idxs]
            if np.any(x < 0) and np.any(x > 0):
                warnings.warn(
                    "You are plotting both n- and p-type carrier concentrations on the "
                    "same figure. Try using the --n-type and --p-type options instead."
                )
            x = np.abs(self.doping[secondary_idxs])
            title_str = fancy_format_temp(self.temperatures[primary_idx])
        else:
            raise ValueError(f"Unrecognised x_property: {x_property}")

        for i, (yi, label) in enumerate(zip(y, labels)):
            label = label.replace("overall", "total")
            if label == "total":
                color = base_total_color if total_color is None else total_color
            else:
                color = f"C{i}"

            if self.average:
                ax.plot(x, yi, label=label, c=color)
            else:
                for j in range(3):
                    ax.plot(
                        x,
                        yi[:, j],
                        label=f"{dir_mapping[j]} {label}",
                        c=color,
                        ls=ls_mapping[j],
                    )

        ylabel = ylabel if ylabel else _property_data["mobility"][self.label_key]
        xlabel = xlabel if xlabel else _property_data[x_property][self.label_key]
        logy = logy if logy is not None else _property_data["mobility"]["log"]
        logx = logx if logx is not None else _property_data[x_property]["log"]
        ylim = get_lim(y, ymin, ymax, logy, self.pad)
        xlim = get_lim(x, xmin, xmax, logx, self.pad)

        if logy:
            ax.semilogy()
        if logx:
            ax.semilogx()

        if not title:
            title_str = None

        ax.set(ylabel=ylabel, xlabel=xlabel, ylim=ylim, xlim=xlim, title=title_str)

    def _get_data(self, n_idx, t_idx):
        if self.separate_mobility:
            labels = sorted(set(self.mechanisms) - {"overall"}) + ["overall"]
            data = np.array([self.get_mobility(m)[n_idx, t_idx] for m in labels])
        else:
            labels = ["mobility"]
            data = self.mobility[n_idx, t_idx][None]  # add dummy first axis

        if self.average:
            data = np.average(np.linalg.eigvalsh(data), axis=-1)
        else:
            data = np.diagonal(data, axis1=-2, axis2=-1)
        return labels, data



import matplotlib.pyplot
from matplotlib import rcParams

from amset.plot.base import (  # noqa
    BaseMeshPlotter,
    BaseTransportPlotter,
    amset_base_style,
    revtex_style,
)

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"


def styled_plot(*style_sheets):
    """Return a decorator that will apply matplotlib style sheets to a plot.

    ``style_sheets`` is a base set of styles, which will be ignored if
    ``no_base_style`` is set in the decorated function arguments.

    The style will further be overwritten by any styles in the ``style``
    optional argument of the decorated function.

    Args:
        style_sheets (:obj:`list`, :obj:`str`, or :obj:`dict`): Any matplotlib
            supported definition of a style sheet. Can be a list of style of
            style sheets.
    """

    def decorator(get_plot):
        def wrapper(*args, fonts=None, style=None, no_base_style=False, **kwargs):

            if no_base_style:
                list_style = []
            else:
                list_style = list(style_sheets)

            if style is not None:
                if "revtex" in style:
                    list_style = [revtex_style]
                elif isinstance(style, list):
                    list_style += style
                else:
                    list_style += [style]

            if fonts is not None:
                list_style += [{"font.family": "sans-serif", "font.sans-serif": fonts}]

            matplotlib.pyplot.style.use(list_style)
            return get_plot(*args, **kwargs)

        return wrapper

    return decorator


def pretty_subplot(
    nrows,
    ncols,
    width=None,
    height=None,
    sharex=False,
    sharey=False,
    dpi=None,
    plt=None,
    gridspec_kw=None,
):
    """Get a :obj:`matplotlib.pyplot` subplot object with pretty defaults.

    Args:
        nrows (int): The number of rows in the subplot.
        ncols (int): The number of columns in the subplot.
        width (:obj:`float`, optional): The width of the plot.
        height (:obj:`float`, optional): The height of the plot.
        sharex (:obj:`bool`, optional): All subplots share the same x-axis.
            Defaults to ``True``.
        sharey (:obj:`bool`, optional): All subplots share the same y-axis.
            Defaults to ``True``.
        dpi (:obj:`int`, optional): The dots-per-inch (pixel density) for
            the plot.
        plt (:obj:`matplotlib.pyplot`, optional): A :obj:`matplotlib.pyplot`
            object to use for plotting.
        gridspec_kw (:obj:`dict`, optional): Gridspec parameters. Please see:
            :obj:`matplotlib.pyplot.subplot` for more information. Defaults
            to ``None``.

    Returns:
        :obj:`matplotlib.pyplot`: A :obj:`matplotlib.pyplot` subplot object
        with publication ready defaults set.
    """

    if width is None:
        width = rcParams["figure.figsize"][0]
    if height is None:
        height = rcParams["figure.figsize"][1]

    # TODO: Make this work if plt is already set...
    if plt is None:
        plt = matplotlib.pyplot
        fig, axes = plt.subplots(
            nrows,
            ncols,
            sharex=sharex,
            sharey=sharey,
            dpi=dpi,
            figsize=(width, height),
            facecolor="w",
            gridspec_kw=gridspec_kw,
        )

    return fig, axes


def get_figsize(nrows, ncols, width=None, height=None, wspace=0.4, hspace=0.3):
    if width is None:
        width = matplotlib.rcParams["figure.figsize"][0]
        width = width * ncols + width * wspace * (ncols - 1)

    if height is None:
        height = matplotlib.rcParams["figure.figsize"][1]
        height = height * nrows + height * hspace * (nrows - 1)

    return width, height
